open Base
open Tree
open Mips

let new_label =
  let c = ref 0 in
    fun prefix -> incr c ; prefix ^ "_" ^ string_of_int !c

(** On utilise la convention suivante: quand une fonction est appelée, ses N
  * arguments sont dans les N dernières "cases" de la pile; ensuite la fonction
  * doit empiler RA et FP; puis elle étend la pile pour réserver une fois pour
  * toute la place pour toutes les variables locales dont on peut avoir besoin;
  * la pile pourra être étendue par la suite, pendant le calcul de la valeur
  * de la fonction, pour stocker les résultats intermédiaires.
  *
  *          x_1
  *          ...
  *          x_n  (les valeurs de nos arguments, passées par l'appelant)
  *          $ra  (adresse où sauter pour retourner dans le code appelant)
  *  $fp --> $fp  (le fp de l'appelé pointe là où on a stocké le fp appelant)
  *          y_1
  *          ...
  *  $sp --> y_m
  *
  * Les cases y_1 .. y_m vont contenir les valeurs de variables locales.
  * Initialement $sp est l'adresse de cette dernière case réservée.
  * Ensuite, pendant le calcul de la fonction, si on a empilé des valeurs
  * de calculs intermédiaires v_0 .. v_k, la pile va ressembler à
  * ceci:
  * 
  *          x_1
  *          ...
  *          x_n
  *          $ra
  *  $fp --> $fp
  *          y_1
  *          ...
  *          y_m
  *          v_0
  *          ...
  *  $sp --> v_k
  *
  * On note que $sp désigne l'adresse de la dernière valeur stockée dans la
  * pile, et non l'adresse de la première case libre dans la pile. *)

(** Table donnant pour chaque variable locale x la position (offset) où elle se
  * trouvera sur la pile par rapport au frame pointer.
  * Attention, on doit aussi y stocker l'info pour les arguments de fonctions. *)
let offsets = Hashtbl.create 50

(** Table donnant pour chaque déclaration (fonction ou variable) le décalage maximal 
   *  (en valeur absolue) pour ;es variables locales utilisées simultanément. 
   * Cela donne la taille  de l'espace à réserver sur la pile. *)
let nvars = Hashtbl.create 10

(** Pour chaque variable x déclarée dans e, calcule la position de la variable
  * dans la pile, par rapport au frame pointer.
  * Renvoie ensuite la position maximale utilisée, correspondant à -4*n où n
  * est le nombre maximal de variables locales simultanément visibles dans e. *)
let rec levels cur e = match e.asa with
  | Cst _ | Var _ -> cur
  | Binop (_,e1,e2) ->
      min (levels cur e1) (levels cur e2)
  | If (c,t,e) ->
      min (levels cur c) (min (levels cur t) (levels cur e))
  | Letin (x,d,b) ->
      Hashtbl.add offsets x (cur-4) ;
      min
        (levels cur d)
        (levels (cur-4) b)
  | Call (f,args) ->
      List.fold_left (fun m e -> min m (levels cur e)) cur args

let preproc_decl d = match d.asa with
  | Vdecl (x,e) ->
      Hashtbl.add nvars x (levels 0 e)
  | Fdecl (f,args,e) ->
      ignore
        (List.fold_right
           (fun x cur ->
              Hashtbl.add offsets x cur ;
              cur+4)
            args 8) ;
      Hashtbl.add nvars f (levels 0 e)

let preproc prog = List.iter preproc_decl prog

let pushr r =
  mips [ Bin(Plus,SP,SP,Oimm(-4)) ;
         Sw(r,Areg(0,SP)) ]
let popr r =
  mips [ Lw(r,Areg(0,SP)) ;
         Bin(Plus,SP,SP,Oimm 4) ]

let rec code_expr r e = match e.asa with
  | Cst (Int i) -> mips [Li (r,i)]
  | Cst (Bool false) -> mips [Li (r,0)]
  | Cst (Bool true) -> mips [Li (r,1)]
  | Var (Global,x) -> mips [Lw (r,Alab x)]
  | Var (Local,x) -> mips [Lw (r,Areg(Hashtbl.find offsets x,FP))]
  | Binop (op,e1,e2) ->
      code_expr r e1 ++
      pushr r ++
      code_expr r e2 ++
      popr T0 ++
      mips [ Bin (op,r,T0,Oreg r) ]
  | Letin (x,e1,e2) ->
      code_expr r e1 ++
      mips [ Sw (r, Areg (Hashtbl.find offsets x, FP)) ] ++
      code_expr r e2
  | Call (f,args) ->
      let evalargs =
        List.fold_left
          (fun evalargs arg ->
             evalargs ++ code_expr r arg ++
             pushr r
             )
          nop args
      in
        evalargs ++
        mips [ Jal f ;
               Bin (Plus,SP,SP,Oimm (4 * List.length args)) ;
               Move (r,V0) ]
  | If (c,t,e) ->
      let lbl_then = new_label "then" in
      let lbl_else = new_label "else" in
      let lbl_end = new_label "end" in
        code_expr r c ++
        mips [ Beqz (r,lbl_else) ;
               Label lbl_then ] ++
        code_expr r t ++
        mips [ B lbl_end ; Label lbl_else ] ++
        code_expr r e ++
        mips [ Label lbl_end ]

(** Préparer la pile pour calculer une fonction en reservant un espace de -n. *)
let initf n = mips [ Bin (Minus,SP,SP,Oimm 8) ;
         Sw (FP,Areg(4,SP)) ;
         Sw (RA,Areg(0,SP)) ;
         Move (FP,SP)] ++
         if n = 0 then nop else mips [ Bin (Plus,SP,SP,Oimm n) ]

(** Nettoyer la pile après le calcul, restaurer RA et FP et return. *)
let finishf n =
  let length = 8 - n in
    mips [ Lw (RA,Areg(0,FP)) ;
           Lw (FP,Areg(4,FP)) ;
           Bin (Plus,SP,SP,Oimm length);
           Jr RA ]

let init x = initf (Hashtbl.find nvars x)
let finish x = finishf (Hashtbl.find nvars x)

(* Pour les varaibles globales, on se contente de positionner FP et de réserver 
    l'espace pour les variables locales *)

let initv x = mips [Move (FP,SP); 
                        Bin (Plus,SP,SP,Oimm (Hashtbl.find nvars x))]  

let finishv = mips [Move (SP,FP)]

(* Print last decl, then bye bye. *)
let display =  mips [ Label "display" ;
             Move (A0,V0) ; Li (V0,1) ; Syscall ;
             Li (A0,10) ; Li (V0,11) ; Syscall ;
             La (A0,"bye") ; Li (V0,4) ; Syscall ]

let code_decls prog =
  let code,defs =
    List.fold_left
      (fun (code,defs) d -> match d.asa with
         | Vdecl (x,e) ->
             code ++
             mips [ Label ("compute_" ^ x) ] 
               ++ initv x ++ code_expr V0 e ++ mips [Sw(V0,Alab x)] ++ finishv,
             defs
         | Fdecl (f,args,e) ->
             code,
             defs ++
             mips [ Label f ] ++  init f ++ code_expr V0 e ++ finish f )
      (nop,nop) prog
  in 
    defs ++ 
    mips [ Label "main" ] ++ initf 0 ++ code ++  display ++ finishf 0 

let globals prog =
  List.fold_left
    (fun globs decl -> match decl.asa with
       | Vdecl (x,_) -> Word (x,0)::globs | _ -> globs)
    [] prog

let generate_program prog =
  preproc prog ;
  { text = code_decls prog;
    data = Asciiz ("bye","Bye\n") :: globals prog }
